// SPDX-License-Identifier: MIT

package gl

import (
	"fmt"
	"reflect"
	"strings"
	"unsafe"
)

// Ptr takes a slice or pointer (to a singular scalar value or the first
// element of an array or slice) and returns its GL-compatible address.
//
// For example:
//
// 	var data []uint8
// 	...
// 	gl.TexImage2D(gl.TEXTURE_2D, ..., gl.UNSIGNED_BYTE, gl.Ptr(&data[0]))
func Ptr(data interface{}) unsafe.Pointer {
	if data == nil {
		return unsafe.Pointer(nil)
	}
	var addr unsafe.Pointer
	switch v := data.(type) {
	case *uint8:
		addr = unsafe.Pointer(v)
	case *uint16:
		addr = unsafe.Pointer(v)
	case *float32:
		addr = unsafe.Pointer(v)
	case []uint8:
		addr = unsafe.Pointer(&v[0])
	case []uint16:
		addr = unsafe.Pointer(&v[0])
	case []float32:
		addr = unsafe.Pointer(&v[0])
	default:
		panic(fmt.Errorf("unsupported type %T; must be a slice or pointer to a singular scalar value or the first element of an array or slice", v))
	}
	return addr
}

// Str takes a null-terminated Go string and returns its GL-compatible address.
// This function reaches into Go string storage in an unsafe way so the caller
// must ensure the string is not garbage collected.
func Str(str string) *uint8 {
	if !strings.HasSuffix(str, "\x00") {
		panic("str argument missing null terminator: " + str)
	}
	header := (*reflect.StringHeader)(unsafe.Pointer(&str))
	return (*uint8)(unsafe.Pointer(header.Data))
}
